:host(.new_chat_button) {
  width: 90%;
  margin-left: 10px;
  margin-top: 10px;
}
:host(.new_chat_button) div button {
  background-color: var(--accent-color) !important;
  text-align: left;
}

:host(.chat_button) {
  width: 90%;
  min-width: 200px;
}

:host(.chat_button) div button {
  overflow: hidden;
  text-overflow: ellipsis;
  text-align: left;
  border: 0px !important;
  font-size: 14px;
  color: rgb(113, 113, 113) !important;
  background-color: unset !important;
}

:host(.chat_button) div button:before {
  content: url("../../imgs/chat_bubble.svg");
  margin-right: 10px;
  display: inline-block;
}

:host(.chat_button.selected) div button,
:host(.chat_button) div button:hover {
  background-color: #f3f3f3 !important;
  border-radius: 0px 5px 5px 0px !important;
  border-left: solid 4px var(--accent-color) !important;
}
