"""
# Web UI

Ragna comes with a web UI where you can try out all of the features!

!!! tip "Under development"

    This documentation page is being actively worked on.
    It will be updated with more details soon. :)
"""

# %%
# You can launch the web application from the CLI:
#
# ```bash
# ragna deploy
# ```
#
# The UI will open in a browser window automatically.
