:host(.main_page_main_row) {
  background-color: rgb(248, 248, 248);
  height: 100%;
  width: 100%;
}

/* Enforces the width of the LeftSidebar 
which is the "first of type" with this class 
(first object in the row) */
:host(.main_page_main_row) .bk-panel-models-layout-Column:first-of-type {
  min-width: 220px;
  width: 15%;
}
